from datetime import datetime, timezone
from typing import Dict, Optional, Union
from uuid import UUID, uuid4
from pydantic import BaseModel
from json import loads
from .sqs_message import SqsMessage
from .functions import parse_message_attributes


class SqsReponse(BaseModel):
    message_id: str
    receipt_handle: str
    md5_of_body: str
    body: str
    attributes: Optional[Dict] = {}
    topic: Optional[str] = ""
    md5_of_message_attributes: Optional[str] = ""
    message_attributes: Optional[Dict] = {}

    def get_payload(self) -> Dict:
        return loads(self.body)

    @staticmethod
    def correlation_id_from_attributes(attributes: Dict) -> Optional[str]:
        correlation_id = attributes.get('correlation_id', None)
        if not correlation_id:
            return None
        if isinstance(correlation_id, str):
            return correlation_id
        if isinstance(correlation_id, dict) and 'Type' in correlation_id and 'Value' in correlation_id:
            return correlation_id["Value"]
        return None

    def get_correlation_id(self) -> Union[str, UUID]:
        correlation_id = self.correlation_id_from_attributes(self.message_attributes)
        if correlation_id:
            return correlation_id
        correlation_id = self.correlation_id_from_attributes(self.attributes)
        if correlation_id:
            return correlation_id
        try:
            s = self.get_payload()['correlation_id']
            return s
        except Exception:
            return str(uuid4())

    def parse_from_sns(self) -> Optional[Dict]:
        payload = self.get_payload()
        if 'Type' not in payload or payload['Type'] != 'Notification':
            return None
        self.topic = str(payload['TopicArn'].split(':')[-1]).lower()
        self.body = payload['Message']
        self.message_attributes = parse_message_attributes(
            payload.get('MessageAttributes', {}))
        return self.get_payload()

    def parse(self) -> Optional[Dict]:
        if self.parse_from_sns() is not None:
            return
        return self.get_payload()

    def get_sqs_message(self) -> SqsMessage:
        sns_payload = self.parse_from_sns()
        if sns_payload:
            if 'correlation_id' not in sns_payload:
                sns_payload['correlation_id'] = str(self.get_correlation_id())
            if 'timestamp' not in sns_payload:
                sns_payload['timestamp'] = datetime.now(timezone.utc).isoformat()
            return SqsMessage(
                id=UUID(sns_payload['id']),
                topic=sns_payload['topic'],
                payload=sns_payload['payload'],
                correlation_id=sns_payload['correlation_id'],
                timestamp=datetime.fromisoformat(sns_payload['timestamp'])
            )
        payload = self.get_payload()
        if 'correlation_id' not in payload:
            payload['correlation_id'] = str(self.get_correlation_id())
        return SqsMessage(
            id=UUID(payload['id']),
            topic=payload['topic'],
            payload=payload['payload'],
            correlation_id=payload['correlation_id'],
            timestamp=datetime.fromisoformat(payload['timestamp'])
        )

    @classmethod
    def from_boto_response(cls, response: Dict):
        return cls(
            message_id=response['MessageId'],
            receipt_handle=response['ReceiptHandle'],
            md5_of_body=response.get('MD5OfBody', ""),
            body=response['Body'],
            attributes=response['Attributes'],
            md5_of_message_attributes=response.get(
                'MD5OfMessageAttributes', ''),
            message_attributes=parse_message_attributes(
                response.get('MessageAttributes', {}))
        )



from typing import Dict


def parse_message_attributes(attributes: Dict) -> Dict:
    message_attributes = {}
    if not attributes:
        return message_attributes
    for key, value in attributes.items():
        if value['Type'] == 'String':
            message_attributes[key] = value['Value']
        elif value['Type'] == 'Number':
            message_attributes[key] = int(value['Value'])
        elif value['Type'] == 'Binary':
            message_attributes[key] = bytes(value['Value'], 'utf-8')
    return message_attributes


def to_message_attributes(attributes: Dict) -> Dict:
    att_dict = {}
    for key, value in attributes.items():
        if isinstance(value, str):
            att_dict[key] = {
                'DataType': 'String', 'StringValue': value}
        elif isinstance(value, bytes):
            att_dict[key] = {
                'DataType': 'Binary', 'BinaryValue': value}
    return att_dict



from datetime import datetime, timezone
from typing import Optional
from uuid import UUID
from pydantic import BaseModel
from json import loads


class SqsMessage(BaseModel):
    id: UUID
    topic: str
    payload: str
    correlation_id: str
    timestamp: Optional[datetime] = datetime.now(timezone.utc)

    def get_payload(self):
        return loads(self.payload)

    def to_dict(self):
        return {
            'id': str(self.id),
            'topic': self.topic,
            'payload': self.payload,
            'correlation_id': self.correlation_id,
            'timestamp': self.timestamp.isoformat()
        }



from finalsa.common.models import (
    SqsMessage, SqsReponse, parse_message_attributes, __version__)
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../../')))


def test_version():
    assert __version__ is not None


def test_parse_message_attributes():
    attributes = {
        'string': {'Type': 'String', 'Value': 'test'},
        'number': {'Type': 'Number', 'Value': '1'},
        'binary': {'Type': 'Binary', 'Value': 'test'}
    }
    result = parse_message_attributes(attributes)
    assert result['string'] == 'test'
    assert result['number'] == 1
    assert result['binary'] == b'test'


def test_sqs_message():
    message = SqsMessage(
        id='123e4567-e89b-12d3-a456-426614174000',
        topic='test',
        payload='{"test": "test"}',
        correlation_id='123e4567-e89b-12d3-a456-426614174000'
    )
    assert str(message.id) == '123e4567-e89b-12d3-a456-426614174000'
    assert message.topic == 'test'
    assert message.get_payload() == {'test': 'test'}
    assert message.correlation_id == '123e4567-e89b-12d3-a456-426614174000'
    assert message.timestamp is not None
    assert message.to_dict() == {
        'id': '123e4567-e89b-12d3-a456-426614174000',
        'topic': 'test',
        'payload': '{"test": "test"}',
        'correlation_id': '123e4567-e89b-12d3-a456-426614174000',
        'timestamp': message.timestamp.isoformat()
    }


def test_sqs_response():
    response = SqsReponse(
        message_id='test',
        receipt_handle='test',
        md5_of_body='test',
        body='test',
        attributes={'test': 'test',
                    'correlation_id': '123e4567-e89b-12d3-a456-426614174000'},
        topic='test',
        md5_of_message_attributes='test',
        message_attributes={'correlation_id': {'Type': 'String',
                                               'Value': '123e4567-e89b-12d3-a456-426614174000'}}

    )
    assert response.body == 'test'
    assert str(response.get_correlation_id()) == '123e4567-e89b-12d3-a456-426614174000'
    assert response.topic == 'test'


def test_sqs_response_from_boto_response():
    boto_response = {
        'MessageId': 'test',
        'ReceiptHandle': 'test',
        'MD5OfBody': 'test',
        'Body': 'test',
        'Attributes': {'test': 'test', 'correlation_id': '123e4567-e89b-12d3-a456-426614174000'},
        'MessageAttributes': {'correlation_id': {'Type': 'String', 'Value': '123e4567-e89b-12d3-a456-426614174000'}}

    }

    response = SqsReponse.from_boto_response(boto_response)
    assert response.body == 'test'
    assert str(response.get_correlation_id()) == '123e4567-e89b-12d3-a456-426614174000'
    assert response.topic == ''
    assert response.message_attributes == {'correlation_id' : "123e4567-e89b-12d3-a456-426614174000"}
    assert response.attributes == {'test': 'test', 'correlation_id': '123e4567-e89b-12d3-a456-426614174000'}


def test_sqs_response_no_correlation_id():
    response = SqsReponse(
        message_id='test',
        receipt_handle='test',
        md5_of_body='test',
        body='test',
        attributes={'test': 'test'},
        topic='test',
        md5_of_message_attributes='test',
    )
    assert response.body == 'test'
    assert response.get_correlation_id() is not None
    assert response.topic == 'test'


def test_sqs_message_from_dict():

    message = SqsMessage(
        id='123e4567-e89b-12d3-a456-426614174000',
        topic='test',
        payload='{"test": "test"}',
        correlation_id='123e4567-e89b-12d3-a456-426614174000'
    )
    message_dict = message.to_dict()
    assert str(message.id) == message_dict['id']
    assert message.topic == message_dict['topic']
    assert message.get_payload() == {'test': 'test'}
    assert message.correlation_id == message_dict['correlation_id']



from finalsa.common.models.models import (
    SqsMessage,
    SqsReponse,
    parse_message_attributes
)

__version__ = "0.0.1"

__all__ = [
    "SqsMessage",
    "SqsReponse",
    "parse_message_attributes"
]
